/* YAD 0.01 (Yet Another Disassembler) (x) herm1t@vx.netlux.org, 2008
 *
 * This also might be called Less eXtended Disassembler Engine, because
 * it's primarily based on XDE v. 1.02 by Z0mbie, the stuff related to
 * operands parsing was omitted, but the table was regenerated to reflect
 * the changes in IA-32 since the last release of XDE. There were also
 * minor clean-ups in the code.
 *
 * The constants and the data structure is partially compatible, if you
 * gonna use this instead of XDE, you'll have to slightly change your code.
 *
 * The code was designed specially for the Linux.Lacrimae virus.
 */
#include "yad.h"

int yad(uint8_t *opcode, yad_t *diza)
{
	uint8_t c, *p;
	uint32_t flags, datasize, addrsize, a, i;

	bzero(diza, sizeof(yad_t));
	p = opcode;	
	datasize = addrsize = 4;
	flags = 0;
	if (*(unsigned short*)p == 0x0000 || *(unsigned short*)p == 0xFFFF)
		flags |= C_BAD;
	/* prefixes */
	for (;;) {
		c = *p++;
#define	GET_PREFIX(cond,field,expr)	\
if (cond) { if (diza->field != 0) flags |= C_BAD; diza->field = c; expr; continue; }
		GET_PREFIX(c == 0x66, p_66, datasize = 2)
		GET_PREFIX(c == 0x67, p_67, addrsize = 2)
		GET_PREFIX((c == 0x26) || (c == 0x2E) || (c == 0x36) || (c == 0x3E) || (c == 0x64) || (c == 0x65), p_seg, )
		GET_PREFIX((c == 0xF2) || (c == 0xF3), p_rep, )
		GET_PREFIX(c == 0xf0, p_lock, )
		break;
	}
	/* first opcode */
	diza->opcode = c;
	/* instruction flags by opcode */
	if (c == 0x0f) {
		/* 2nd opcode */
		diza->opcode2 = c = *p++;
		flags |= yad_table[256 + c];
	} else
		flags |= yad_table[c];
	if (flags == C_ERROR)
		return 0;
	if (flags & C_PREGRP) {
		i = diza->p_rep == 0xf2 ? 3 :
			(diza->p_66 ? 2 : (diza->p_rep == 0xf3 ? 1 : 0));
		flags = yad_table[(flags & YAD_GROUP_MASK) + i];
	}
	/* parse ModRM and SIB */
	if (flags & (C_MODRM|C_GROUP)) {
		uint32_t mod, rm;
		c = *p++;
		diza->modrm = c;
		mod = c >> 6;
		rm  = c & 7;
		if (flags & C_GROUP)
			flags = yad_table[(flags & YAD_GROUP_MASK) + ((c >> 3) & 7)];
		if (mod != 3) {
			if (rm == 4) {
				c = *p++;
				flags |= C_SIB;
				diza->sib = c;
			}
			if (mod == 2 || (mod == 0 && (rm == 5 || (rm == 4 && (c & 7) == 5))))
				flags |= C_ADDR4;
			else if (mod == 1)
				flags |= C_ADDR1;
		}
	}
	diza->flags = flags;
#define	COPY_ARG(size,flag,var,save)	\
a = size;				\
if (flags & flag)			\
	a += var;			\
for (i = 0; i < a; i++)			\
	*(&save + i) = *p++;		\
diza->var = a;
	COPY_ARG(flags & 7, C_ADDR67, addrsize, diza->addr1);
	COPY_ARG((flags >> 7) & 7, C_DATA66, datasize, diza->data1);
	if ((diza->len = p - opcode) > 15)
		return 0;	
	return diza->len;
}

int yad_asm(uint8_t *opcode, yad_t *diza)
{
	uint8_t *p;
	int i;

	p = opcode;
	if (diza->p_seg)
		*p++ = diza->p_seg;
	if (diza->p_lock)
		*p++ = diza->p_lock;
	if (diza->p_rep)
		*p++ = diza->p_rep;
	if (diza->p_67)
		*p++ = diza->p_67;
	if (diza->p_66)
		*p++ = diza->p_66;
	*p++ = diza->opcode;
	if (diza->opcode == 0x0F)
		*p++ = diza->opcode2;
	if (diza->flags & C_MODRM)
		*p++ = diza->modrm;
	if (diza->flags & C_SIB)
		*p++ = diza->sib;
	for (i = 0; i < diza->addrsize; i++)
		*p++ = *(&diza->addr1 + i);
	for (i = 0; i < diza->datasize; i++)
		*p++ = *(&diza->data1 + i);
	return p - opcode;
}
